# -*- mode: python; coding: utf-8 -*-
# Copyright 2017-2018 Peter Williams and collaborators.
# Licensed under the MIT License.

"""The main neural network code.

"""
from __future__ import absolute_import, division, print_function

__all__ = '''
NSModel
PhysicalApproximator
'''.split()

import os.path
from six.moves import range
import numpy as np
from pwkit import cgs
from pwkit.numutil import broadcastize
from keras import models, layers, optimizers

from . import DomainRange

hardcoded_nu_ref = 1.0
hardcoded_ne_ref = 1.0


class NSModel(models.Sequential):
    """Neuro-Synchro Model -- just keras.models.Sequential extended with some
    helpers specific to our data structures. If you run the `ns_setup` method
    you can train the neural net in our system.

    """
    # NOTE: __init__() must take no arguments in order for keras to be able to
    # deserialize NSModels from the HDF5 format.

    def ns_setup(self, result_index, data):
        self.result_index = int(result_index)
        self.domain_range = data.domain_range
        self.data = data
        assert self.result_index < self.domain_range.n_results
        return self # convenience


    def ns_fit(self, **kwargs):
        """Train this ANN model on the data in `self.data`. This function just
        takes care of extracting the right parameter and avoiding NaNs.

        """
        nres = self.data.norm_results[:,self.result_index]
        ok = np.isfinite(nres)
        nres = nres[ok].reshape((-1, 1))
        npar = self.data.norm_params[ok]
        return self.fit(npar, nres, **kwargs)


    def ns_validate(self, filter=True, to_phys=True):
        """Test this network by having it predict all of the values in our training
        sample. Returns `(params, actual, nn)`, where `params` is shape `(N,
        self.data.n_params)` and is the input parameters, `actual` is shape
        `(N,)` and is the actual values returned by the calculator, and `nn`
        is shape `(N,)` and is the values predicted by the neural net.

        If `filter` is true, the results will be filtered such that neither
        `actual` nor `nn` contain non-finite values.

        If `to_phys` is true, the values will be returned in the physical
        coordinate system. Otherwise they will be returned in the normalized
        coordinate system.

        """
        if to_phys:
            par = self.data.phys_params
            res = self.data.phys_results[:,self.result_index]
        else:
            par = self.data.norm_params
            res = self.data.norm_results[:,self.result_index]

        npred = self.predict(self.data.norm_params)[:,0]

        if filter:
            ok = np.isfinite(res) & np.isfinite(npred)
            par = par[ok]
            res = res[ok]
            npred = npred[ok]

        if to_phys:
            pred, _ = self.domain_range.rmaps[self.result_index].norm_to_phys(npred)
        else:
            pred = npred

        return par, res, pred


    def ns_sigma_clip(self, n_norm_sigma):
        """Assuming that self is already a decent approximation of the input data, try
        to improve things by NaN-ing out any measurements that are extremely
        discrepant with our approximation -- under the assumption that these
        are cases where the calculator went haywire.

        Note that this destructively modifies `self.data`.

        `n_norm_sigma` is the threshold above which discrepant values are
        flagged. It is evaluated using the differences between the neural net
        prediction and the training data in the *normalized* coordinate
        system.

        Returns the number of flagged points.

        """
        nres = self.data.norm_results[:,self.result_index]
        npred = self.predict(self.data.norm_params)[:,0]
        err = npred - nres
        m = np.nanmean(err)
        s = np.nanstd(err)
        bad = (np.abs((err - m) / s) > n_norm_sigma)
        self.data.phys[bad,self.domain_range.n_params+self.result_index] = np.nan
        self.data.norm[bad,self.domain_range.n_params+self.result_index] = np.nan
        return bad.sum()


    def ns_plot(self, param_index, plot_err=False, to_phys=False, thin=100):
        """Make a diagnostic plot comparing the approximation to the "actual" results
        from the calculator.

        """
        import omega as om

        par, act, nn = self.ns_validate(filter=True, to_phys=to_phys)

        if plot_err:
            err = nn - act
            p = om.quickXY(par[::thin,param_index], err[::thin], 'Error', lines=0)
        else:
            p = om.quickXY(par[::thin,param_index], act[::thin], 'Full calc', lines=0)
            p.addXY(par[::thin,param_index], nn[::thin], 'Neural', lines=0)

        return p


class PhysicalApproximator(object):
    """This class approximates the eight nontrivial RT coefficients using a
    physically-based parameterization.

    See :ref:`how-to-use` for detailed documentation of how to prepare and
    train the neural networks used by this class.

    **Constructor argument**

    **nn_dir**
      The path to a directory containing trained neural network data. This
      directory should contain the configuration file ``nn_config.toml`` and
      serialized neural network weights in files with names like
      ``rho_Q_sign.h5``.

    """
    results = 'j_I alpha_I rho_Q rho_V j_frac_pol alpha_frac_pol j_V_share alpha_V_share rho_Q_sign'.split()

    def __init__(self, nn_dir):
        self.domain_range = DomainRange.from_serialized(os.path.join(nn_dir, 'nn_config.toml'))

        for i, r in enumerate(self.results):
            assert self.domain_range.rmaps[i].name == r

            m = models.load_model(
                os.path.join(nn_dir, '%s.h5' % r),
                custom_objects = {'NSModel': NSModel}
            )
            m.result_index = i
            m.domain_range = self.domain_range
            setattr(self, r, m)


    @broadcastize(4, ret_spec=None)
    def compute_all_nontrivial(self, nu, B, n_e, theta, **kwargs):
        """Compute the nontrivial radiative transfer coefficients.

        **Arguments**

        *nu*
          The observing frequency, in Hz. This and all parameters may be
          scalars or arrays; they are broadcast to a common shape before
          performing the computations.
        *B*
          The local magnetic field strength, in G.
        *n_e*
          The local density of synchrotron-emitting particles, in cm^-3.
        *theta*
          The angle between the line of sight and the local magnetic field,
          in radians.
        ``**kwargs``
          Other arguments to the synchrotron model; these can vary depending
          on which particle distribution was used.

        **Return values**

        A tuple of ``(coeffs, oos)``:

        *coeffs*
          The radiative transfer coefficients in the Stokes basis where the
          Stokes U axis is aligned with the magnetic field. This is an array
          of shape ``S + (8,)`` where *S* is the shape of the broadcasted
          input parameters. Along the inner axis of the array, the coefficients
          are: ``(j_I, alpha_I, j_Q, alpha_Q, j_V, alpha_V, rho_Q, rho_V)``.
        *oos*
          An array of integers reporting where the calculations encountered
          out-of-sample values, that is, inputs or outputs beyond the range in
          which the neural networks were trained. The shape of this array is
          the same as that of the broadcased input parameters, or a scalar if
          the inputs were all scalars. For each set of input parameters, the
          least significant bit is set to 1 if the first input parameter was
          out-of-sample, where "first" is defined by the order in which these
          parameters are listed in the ``nn_config.toml`` file. The next more
          significant bit is set if the second input parameter was out of
          sample, and so on. After all of the input parameters, there are 9
          flag bits indicating whether any of the *output* results were
          out-of-sample, relative to the range of normalized values
          encountered in the training set. The order in which these parameters
          are processed is ``j_I``, ``alpha_I``, ``rho_Q``, ``rho_V``,
          ``j_frac_pol``, ``alpha_frac_pol``, ``j_V_share``,
          ``alpha_V_share``, ``rho_Q_sign``. Therefore if the synchrotron
          model takes 4 input parameters and the ``rho_Q_sign`` output is the
          only one to have been out-of-sample, the resulting ``oos`` value
          will be ``0x1000``.

        """
        # Turn the standard parameters into the ones used in our computations

        no_B = np.logical_not(B > 0)
        nu_cyc = cgs.e * B / (2 * np.pi * cgs.me * cgs.c)
        nu_cyc[no_B] = 1e7 # fake to avoid div-by-0 for now
        kwargs['s'] = nu / nu_cyc

        # Normalize theta (assuming it could take on any value

        theta = theta % (2 * np.pi)
        w = (theta > np.pi)
        theta[w] = 2 * np.pi - theta[w]
        flip = (theta > 0.5 * np.pi)
        theta[flip] = np.pi - theta[flip]
        kwargs['theta'] = theta

        # Normalize inputs and check for out-of-sample.

        oos_flags = 0

        norm = np.empty(nu.shape + (self.domain_range.n_params,))
        for i, mapping in enumerate(self.domain_range.pmaps):
            norm[...,i], flag = mapping.phys_to_norm(kwargs[mapping.name])
            if flag:
                oos_flags |= (1 << i)

        # Compute base outputs.

        j_I, flag = self.domain_range.rmaps[0].norm_to_phys(self.j_I.predict(norm)[...,0])
        if flag:
            oos_flags |= (1 << (self.domain_range.n_params + 0))

        alpha_I, flag = self.domain_range.rmaps[1].norm_to_phys(self.alpha_I.predict(norm)[...,0])
        if flag:
            oos_flags |= (1 << (self.domain_range.n_params + 1))

        rho_Q, flag = self.domain_range.rmaps[2].norm_to_phys(self.rho_Q.predict(norm)[...,0])
        if flag:
            oos_flags |= (1 << (self.domain_range.n_params + 2))

        rho_V, flag = self.domain_range.rmaps[3].norm_to_phys(self.rho_V.predict(norm)[...,0])
        if flag:
            oos_flags |= (1 << (self.domain_range.n_params + 3))

        j_frac_pol, flag = self.domain_range.rmaps[4].norm_to_phys(self.j_frac_pol.predict(norm)[...,0])
        if flag:
            oos_flags |= (1 << (self.domain_range.n_params + 4))

        alpha_frac_pol, flag = self.domain_range.rmaps[5].norm_to_phys(self.alpha_frac_pol.predict(norm)[...,0])
        if flag:
            oos_flags |= (1 << (self.domain_range.n_params + 5))

        j_V_share, flag = self.domain_range.rmaps[6].norm_to_phys(self.j_V_share.predict(norm)[...,0])
        if flag:
            oos_flags |= (1 << (self.domain_range.n_params + 6))

        alpha_V_share, flag = self.domain_range.rmaps[7].norm_to_phys(self.alpha_V_share.predict(norm)[...,0])
        if flag:
            oos_flags |= (1 << (self.domain_range.n_params + 7))

        rho_Q_sign, flag = self.domain_range.rmaps[8].norm_to_phys(self.rho_Q_sign.predict(norm)[...,0])
        if flag:
            oos_flags |= (1 << (self.domain_range.n_params + 8))

        # Patch up B = 0 in the obvious way. (Although if we ever have to deal
        # with nontrivial cold plasma densities, zones of zero B might affect
        # the RT if they cause refraction or what-have-you.)

        j_I[no_B] = 0.
        alpha_I[no_B] = 0.

        # Un-transform, baking in the invariant that our Q parameters are
        # always negative and the V parameters are always positive (given our
        # theta normalization).

        j_P = j_frac_pol * j_I
        j_V = j_V_share * j_P
        j_Q = -np.sqrt(1 - j_V_share**2) * j_P

        alpha_P = alpha_frac_pol * alpha_I
        alpha_V = alpha_V_share * alpha_P
        alpha_Q = -np.sqrt(1 - alpha_V_share**2) * alpha_P

        rho_Q = rho_Q * rho_Q_sign

        # Now apply the known scalings.

        n_e_scale = n_e / hardcoded_ne_ref
        j_I *= n_e_scale
        alpha_I *= n_e_scale
        j_Q *= n_e_scale
        alpha_Q *= n_e_scale
        j_V *= n_e_scale
        alpha_V *= n_e_scale
        rho_Q *= n_e_scale
        rho_V *= n_e_scale

        freq_scale = nu / hardcoded_nu_ref
        j_I *= freq_scale
        alpha_I /= freq_scale
        j_Q *= freq_scale
        alpha_Q /= freq_scale
        j_V *= freq_scale
        alpha_V /= freq_scale
        rho_Q /= freq_scale
        rho_V /= freq_scale

        theta_sign_term = np.ones(n_e.shape, dtype=np.int)
        theta_sign_term[flip] = -1
        j_V *= theta_sign_term
        alpha_V *= theta_sign_term
        rho_V *= theta_sign_term

        # Pack it up and we're done.

        result = np.empty(n_e.shape + (8,))
        result[...,0] = j_I
        result[...,1] = alpha_I
        result[...,2] = j_Q
        result[...,3] = alpha_Q
        result[...,4] = j_V
        result[...,5] = alpha_V
        result[...,6] = rho_Q
        result[...,7] = rho_V
        return result, oos_flags



# -*- mode: python; coding: utf-8 -*-
# Copyright 2017-2018 Peter Williams and collaborators.
# Licensed under the MIT License.

"""Train and use artificial neural network approximations ("regressions") of
synchrotron radiative transfer coefficients as a function of various physical
input parameters.

The meat of the neural network code is in the ``impl`` module to avoid
importing Keras unless needed.

"""
from __future__ import absolute_import, division, print_function

__all__ = '''
AbsLogMapping
DirectMapping
DomainRange
LogMapping
LogitMapping
Mapping
NegLogMapping
NinthRootMapping
SampleData
SignMapping
basic_load
detrivialize_stokes_basis
mapping_from_dict
mapping_from_samples
'''.split()

from collections import OrderedDict
from six.moves import range
import numpy as np
import pandas as pd
from pwkit.io import Path


class Mapping(object):
    """An abstract base class for parameter transformations.

    Mapping classes are used to translate the physical parameters fed into
    *neurosynchro* into normalized values that are easier to work with
    numerically. You will not normally need to use them directly.

    """
    desc = None # set by subclasses
    trainer = None

    phys_bounds_mode = 'empirical'
    "Blah blah"

    normalization_mode = 'gaussian'
    out_of_sample = 'ignore'

    def __init__(self, name):
        self.name = name


    @classmethod
    def from_info_and_samples(cls, info, phys_samples):
        """Create a new :class:`Mapping` from a dictionary of information and a set of samples.

        **Call signature**

        *info*
          A dictionary of attributes, passed into :meth:`Mapping.from_dict`.
        *phys_samples*
          A 1D Numpy array of samples of this parameter, in no particular order.
        Return value
          A new instance of :class:`Mapping` (or one of its subclasses) with initialized
          bounds parameters.

        """
        inst = cls.from_dict(info, load_bounds=False)

        valid = np.isfinite(phys_samples) & inst._is_valid(phys_samples)
        n_rej = phys_samples.size - valid.sum()
        print('%s: rejecting %d samples out of %d' % (inst.name, n_rej, phys_samples.size))
        phys_samples = phys_samples[valid]
        if phys_samples.size < 3:
            raise Exception('not enough valid samples for %s' % inst.name)

        if inst.phys_bounds_mode == 'empirical':
            inst.p_min = phys_samples.min()
            inst.p_max = phys_samples.max()
        elif inst.phys_bounds_mode == 'theta':
            inst.p_min = 0.
            inst.p_max = 0.5 * np.pi
        else:
            raise ValueError('unrecognized phys_bounds_mode value %r for %s' %
                             (inst.phys_bounds_mode, inst.name))

        # Pluggable "transform"
        transformed = inst._to_xform(phys_samples)

        if inst.normalization_mode == 'gaussian':
            inst.x_mean = transformed.mean()
            inst.x_std = transformed.std()
        elif inst.normalization_mode == 'unit_interval':
            # Maps the physical values to the unit interval [0, 1].
            inst.x_mean = transformed.min()
            inst.x_std = transformed.max() - inst.x_mean
            if inst.x_std == 0:
                inst.x_std = 1.
        else:
            raise ValueError('unrecognized normalization_mode value %r for %s' %
                             (inst.normalization_mode, inst.name))

        # Normalize
        normed = (transformed - inst.x_mean) / inst.x_std
        inst.n_min = normed.min()
        inst.n_max = normed.max()

        return inst


    def __repr__(self):
        return '<Mapping %s %s mean=%r sd=%r>' % (self.name, self.desc, self.x_mean, self.x_std)


    def phys_to_norm(self, phys):
        """Map "physical" parameters to normalized values

        **Argument**

        *phys*
          An array of "physical" input values (see :ref:`transformations`).

        **Return values**

        This method returns a tuple ``(normalized, oos)``.

        *normalized*
          The normalized versions of the input data.
        *oos*
          An array of booleans of the same shape as the input data. True
          values indicate inputs that were out of the sample that was used to
          define the mapping.

        """
        # note: using prefix ~ instead of np.logical_not fails for scalars
        oos = np.logical_not((phys >= self.p_min) & (phys <= self.p_max)) # catches NaNs
        any_oos = np.any(oos)

        if any_oos:
            if self.out_of_sample == 'ignore':
                pass
            elif self.out_of_sample == 'clip':
                phys = np.clip(phys, self.p_min, self.p_max)
            elif self.out_of_sample == 'nan':
                phys = phys.copy()
                phys[oos] = np.nan
            else:
                raise Exception('unrecognized out-of-sample behavior %r' % self.out_of_sample)

        return (self._to_xform(phys) - self.x_mean) / self.x_std, any_oos


    def norm_to_phys(self, norm):
        """Map "normalized" parameters to "physical" values

        **Argument**

        *norm*
          An array of "normalized" input values (see :ref:`transformations`).

        **Return values**

        This method returns a tuple ``(phys, oos)``.

        *phys*
          The physical versions of the input data.
        *oos*
          An array of booleans of the same shape as the input data. True
          values indicate inputs that were out of the sample that was used to
          define the mapping.

        """
        oos = np.logical_not((norm >= self.n_min) & (norm <= self.n_max)) # catches NaNs
        any_oos = np.any(oos)

        if any_oos:
            if self.out_of_sample == 'ignore':
                pass
            elif self.out_of_sample == 'clip':
                norm = np.clip(norm, self.n_min, self.n_max)
            elif self.out_of_sample == 'nan':
                norm = norm.copy()
                norm[oos] = np.nan
            else:
                raise Exception('unrecognized out-of-sample behavior %r' % self.out_of_sample)

        return self._from_xform(norm * self.x_std + self.x_mean), any_oos


    def to_dict(self):
        """Serialize this :class:`Mapping` into an ordered dictionary."""
        d = OrderedDict()
        d['name'] = self.name
        d['maptype'] = self.desc

        if self.phys_bounds_mode is not None:
            d['phys_bounds_mode'] = self.phys_bounds_mode
        if self.normalization_mode is not None:
            d['normalization_mode'] = self.normalization_mode
        if self.trainer is not None:
            d['trainer'] = self.trainer
        if self.out_of_sample is not None:
            d['out_of_sample'] = self.out_of_sample

        d['x_mean'] = self.x_mean
        d['x_std'] = self.x_std
        d['phys_min'] = self.p_min
        d['phys_max'] = self.p_max
        d['norm_min'] = self.n_min
        d['norm_max'] = self.n_max
        return d


    @classmethod
    def from_dict(cls, info, load_bounds=True):
        """Deserialize an ordered dictionary into a new :class:`Mapping` instance.

        **Call signature**

        *info*
          A dictionary of parameters, as generated by :meth:`Mapping.to_dict`.
        *load_bounds* (default: :const:`True`)
          If true, deserialize bounds information such as the maximum and minimum
          observed physical values. If :const:`False`, these are left uninitialized.
        Return value
          A new :class:`Mapping` instance.

        """
        if str(info['maptype']) != cls.desc:
            raise ValueError('info is for maptype %s but this class is %s' % (info['maptype'], cls.desc))

        inst = cls(str(info['name']))
        if 'phys_bounds_mode' in info:
            inst.phys_bounds_mode = info['phys_bounds_mode']
        if 'normalization_mode' in info:
            inst.normalization_mode = info['normalization_mode']
        if 'trainer' in info:
            inst.trainer = info['trainer']
        if 'out_of_sample' in info:
            inst.out_of_sample = info['out_of_sample']

        if load_bounds:
            inst.x_mean = float(info['x_mean'])
            inst.x_std = float(info['x_std'])
            inst.p_min = float(info['phys_min'])
            inst.p_max = float(info['phys_max'])
            inst.n_min = float(info['norm_min'])
            inst.n_max = float(info['norm_max'])

        return inst


class AbsLogMapping(Mapping):
    desc = 'abs_log'

    def _to_xform(self, p):
        return np.log10(np.abs(p))

    def _from_xform(self, x):
        return 10**x # XXX not invertible!

    def _is_valid(self, p):
        return p != 0


class DirectMapping(Mapping):
    desc = 'direct'

    def _to_xform(self, p):
        return p

    def _from_xform(self, x):
        return x

    def _is_valid(self, p):
        return np.ones(p.shape, dtype=np.bool)


class LogMapping(Mapping):
    desc = 'log'

    def _to_xform(self, p):
        return np.log10(p)

    def _from_xform(self, x):
        return 10**x

    def _is_valid(self, p):
        return (p > 0)


class LogitMapping(Mapping):
    desc = 'logit'

    def _to_xform(self, p):
        return np.log(p / (1. - p))

    def _from_xform(self, x):
        return np.exp(x) / (np.exp(x) + 1)

    def _is_valid(self, p):
        # Infinities are hard to deal with so we don't allow p = 0 or p = 1.
        return (p > 0) & (p < 1)


class NegLogMapping(Mapping):
    desc = 'neg_log'

    def _to_xform(self, p):
        return np.log10(-p)

    def _from_xform(self, x):
        return -(10**x)

    def _is_valid(self, p):
        return (p < 0)


class NinthRootMapping(Mapping):
    desc = 'ninth_root'

    def _to_xform(self, p):
        return np.cbrt(np.cbrt(p))

    def _from_xform(self, x):
        return x**9

    def _is_valid(self, p):
        return np.ones(p.shape, dtype=np.bool)


class SignMapping(Mapping):
    desc = 'sign'

    def _to_xform(self, p):
        return np.sign(p)

    def _from_xform(self, x):
        return np.sign(x) # XXX not reversible!

    def _is_valid(self, p):
        return (p != 0.)


_mappings = {
    'abs_log': AbsLogMapping,
    'direct': DirectMapping,
    'log': LogMapping,
    'logit': LogitMapping,
    'neg_log': NegLogMapping,
    'ninth_root': NinthRootMapping,
    'sign': SignMapping,
}


class Passthrough(object):
    def __init__(self, d):
        self.d = d

    def to_dict(self):
        return self.d


def mapping_from_info_and_samples(info, phys_samples):
    """Create a :class:`Mapping` subclass from configuration data and sample data.

    **Call signature**

    *info*
      A dictionary of configuration info, loaded from ``nn_config.toml`` or
      created by :meth:`Mapping.to_dict`.
    *phys_samples*
      A 1D array of training data used to initialize the bounds of this mapping.
    Return value
      A new :class:`Mapping` instance. The particular subclass used depends on the
      ``maptype`` setting, as documented in the :ref:`map-types`.

    """
    cls = _mappings[info['maptype']]
    return cls.from_info_and_samples(info, phys_samples)

def mapping_from_dict(info):
    """Create a :class:`Mapping` subclass from configuration data.

    **Call signature**

    *info*
      A dictionary of configuration info, loaded from ``nn_config.toml`` or
      created by :meth:`Mapping.to_dict`.
    Return value
      A new :class:`Mapping` instance. The particular subclass used depends on the
      ``maptype`` setting, as documented in the :ref:`map-types`.

    Unlike :func:`mapping_from_info_and_samples`, the bounds data are not initialized
    if they are not already defined in the configuration dictionary.

    """
    maptype = str(info['maptype'])
    cls = _mappings[maptype]
    return cls.from_dict(info)


def basic_load(datadir, drop_metadata=True):
    """Load a directory of textual tables (such as training set data).

    **Call signature**

    *datadir*
      A path to a directory of textual tables; format described below.
    *drop_metadata* (default ``True``)
      If true, columns marked as metadata will be dropped from the returned
      table.
    Return value
      A :class:`pandas.DataFrame` of data, concatenating all of the input tables.

    The data format is documented in :ref:`make-training-set`. Briefly, each
    file in *datadir* whose name ends with ``.txt`` will be loaded as a table
    using :func:`pandas.read_table`. The recommended format is tab-separated
    values with a single header row. Column names should end in type
    identifiers such as ``(lin)`` to identify their roles, although this
    function ignores this information except to drop columns whose names end
    in ``(meta)`` if so directed.

    """
    datadir = Path(datadir)
    chunks = []

    for item in datadir.glob('*.txt'):
        chunks.append(pd.read_table(str(item)))

    data = pd.concat(chunks, ignore_index=True)

    if drop_metadata:
        # Drop `foo(mtea)` columns
        for col in data.columns:
            if col.endswith('(meta)'):
                del data[col]

    return data


class DomainRange(object):
    n_params = None
    n_results = None
    pmaps = None
    rmaps = None


    @classmethod
    def from_info_and_samples(cls, info, df):
        inst = cls()
        inst.n_params = len(info['params'])
        inst.n_results = len(info['results'])
        inst.pmaps = []
        inst.rmaps = []

        for i, pinfo in enumerate(info['params']):
            if not df.columns[i].startswith(pinfo['name'] + '('):
                raise Exception('alignment error: expect data column %d to be %r; got %r' % (i, pinfo['name'], df.columns[i]))
            inst.pmaps.append(mapping_from_info_and_samples(pinfo, df[df.columns[i]].values))

        n = inst.n_params

        for i, rinfo in enumerate(info['results']):
            if not df.columns[n+i].startswith(rinfo['name'] + '(res)'):
                raise Exception('alignment error: expect data column %d to be %r; got %r' % (n + i, rinfo['name'], df.columns[n + i]))
            inst.rmaps.append(mapping_from_info_and_samples(rinfo, df[df.columns[n+i]].values))

        return inst


    @classmethod
    def from_serialized(cls, config_path, result_to_extract=None):
        """`result_to_extract` is a total lazy hack for the training tool."""
        import pytoml

        with Path(config_path).open('rt') as f:
            info = pytoml.load(f)

        inst = cls()
        inst.pmaps = []
        inst.rmaps = []
        extracted_info = None

        for subinfo in info['params']:
            inst.pmaps.append(mapping_from_dict(subinfo))

        for i, subinfo in enumerate(info['results']):
            if result_to_extract is not None and subinfo['name'] == result_to_extract:
                extracted_info = subinfo
                extracted_info['_index'] = i
            inst.rmaps.append(mapping_from_dict(subinfo))

        inst.n_params = len(inst.pmaps)
        inst.n_results = len(inst.rmaps)

        if result_to_extract is not None:
            return inst, extracted_info
        return inst


    def __repr__(self):
        return '\n'.join(
            ['<%s n_p=%d n_r=%d' % (self.__class__.__name__, self.n_params, self.n_results)] +
            ['  P%d=%r,' % (i, m) for i, m in enumerate(self.pmaps)] +
            ['  R%d=%r,' % (i, m) for i, m in enumerate(self.rmaps)] +
            ['>'])


    def into_info(self, info):
        info['params'] = [m.to_dict() for m in self.pmaps]
        info['results'] = [m.to_dict() for m in self.rmaps]


    def load_and_normalize(self, datadir):
        df = basic_load(datadir)

        for i, pmap in enumerate(self.pmaps):
            if not df.columns[i].startswith(pmap.name + '('):
                raise Exception('alignment error: expect data column %d to be %s(...); got %s' % (i, pmap.name, df.columns[i]))

        n = self.n_params

        for i, rmap in enumerate(self.rmaps):
            if df.columns[n+i] != rmap.name + '(res)':
                raise Exception('alignment error: expect data column %d to be %s(res); got %s' % (n + i, rmap.name, df.columns[n + i]))

        return SampleData(self, df)


class SampleData(object):
    df = None
    domain_range = None
    phys = None
    norm = None
    oos_flags = 0

    def __init__(self, domain_range, df):
        self.df = df
        self.domain_range = domain_range
        self.phys = df.values
        self.norm = np.empty_like(self.phys)

        for i in range(self.domain_range.n_params):
            self.norm[:,i], flag = self.domain_range.pmaps[i].phys_to_norm(self.phys[:,i])
            if flag:
                self.oos_flags |= (1 << i)

        for i in range(self.domain_range.n_results):
            j = i + self.domain_range.n_params
            self.norm[:,j], flag = self.domain_range.rmaps[i].phys_to_norm(self.phys[:,j])
            if flag:
                self.oos_flags |= (1 << j)

    @property
    def phys_params(self):
        return self.phys[:,:self.domain_range.n_params]

    @property
    def phys_results(self):
        return self.phys[:,self.domain_range.n_params:]

    @property
    def norm_params(self):
        return self.norm[:,:self.domain_range.n_params]

    @property
    def norm_results(self):
        return self.norm[:,self.domain_range.n_params:]


def detrivialize_stokes_basis(coeffs, psi):
    """Re-express coefficients in a basis in which the magnetic field can rotate
    on the sky.

    **Arguments**

    *coeffs*
      Radiative transfer coefficients in the Stokes basis where the Stokes U
      axis is aligned with the magnetic field. This is an array of shape ``(S,
      8)`` where *S* is the shape of *psi*. Along the inner axis of the array,
      the coefficients are: ``(j_I, alpha_I, j_Q, alpha_Q, j_V, alpha_V,
      rho_Q, rho_V)``. This is the representation returned by
      :meth:`neurosynchro.impl.PhysicalApproximator.compute_all_nontrivial`.

    *psi*
      The angle(s) between the magnetic field as projected on the sky and some
      invariant Stokes U axis, in radians. XXX: sign convention?

    **Return value**

    An array of radiative transfer coefficients in which the Stokes U terms
    are no longer trivial. The shape is ``(S, 11)``. Along the inner axis of
    the array, the coefficients are: ``(j_I, alpha_I, j_Q, alpha_Q, j_U,
    alpha_U, j_V, alpha_V, rho_Q, rho_U, rho_V)``.

    **Details**

    Synchrotron calculations are generally performed in a basis in which the
    Stokes U axis is aligned with the magnetic field, which means that the
    corresponding radiative transfer coefficients are zero ("trivial"). In
    actual work, however, the magnetic field orientation is not guaranteed to
    be constant along the direction of propagation. If the Stokes U axis is
    held fixed along the integration, the Stokes U coefficients become
    nontrivial. This function transforms coefficients from the former basis
    to the latter.

    See Shcherbakov & Huang (2011MNRAS.410.1052S), equations 50-51. Note
    that the linear polarization axis rotates at twice the rate that psi
    does, because linear polarization is an *orientation* not an *angle*.

    """
    ji = coeffs[...,0]
    ai = coeffs[...,1]
    jq = coeffs[...,2]
    aq = coeffs[...,3]
    jv = coeffs[...,4]
    av = coeffs[...,5]
    rq = coeffs[...,6]
    rv = coeffs[...,7]

    twochi = 2 * (np.pi - psi) # note the sign convention!
    s = np.sin(twochi)
    c = np.cos(twochi)

    xformed = np.empty(coeffs.shape[:-1] + (11,))
    xformed[...,0] = ji # j_I
    xformed[...,1] = ai # alpha_I
    xformed[...,2] = c * jq # j_Q
    xformed[...,3] = c * aq # alpha_Q
    xformed[...,4] = s * jq # j_U
    xformed[...,5] = s * aq # alpha_U
    xformed[...,6] = jv # j_V
    xformed[...,7] = av # alpha_V
    xformed[...,8] = c * rq # rho_Q
    xformed[...,9] = s * rq # rho_U
    xformed[...,10] = rv # rho_V

    return xformed



# -*- mode: python; coding: utf-8 -*-
# Copyright 2018 Peter Williams and collaborators.
# Licensed under the MIT License.

"""This module provides helpers for doing radiative transfer integrations with
`grtrans <https://github.com/jadexter/grtrans>`_, including a simple
framework for running end-to-end tests.

In order to use this functionality, the Python module ``radtrans_integrate``
must be importable. Sadly `grtrans <https://github.com/jadexter/grtrans>`_
doesn't install itself like a regular Python package, so getting this working
can be a pain. Documenting the installation procedure is beyond the scope of
this project.

"""
from __future__ import absolute_import, division, print_function

import numpy as np
import pandas as pd


# grtrans' integration methods:
METHOD_LSODA_YES_LINEAR_STOKES = 0 # LSODA with IS_LINEAR_STOKES=1
METHOD_DELO = 1 # DELO method from Rees+ (1989ApJ...339.1093R)
METHOD_FORMAL = 2 # "formal" method = "matricant (O-matrix) method from Landi Degl'Innocenti"
METHOD_LSODA_NO_LINEAR_STOKES = 3 # LSODA with IS_LINEAR_STOKES=0 -- this is "under development" spherical stokes


def integrate_ray_formal(x, j, K):
    """Use `grtrans <https://github.com/jadexter/grtrans>`_ to integrate one ray
    using its "formal" (matricant / O-matrix) method.

    **Call signature**

    *x*
      1D array, shape (n,). Path length along ray, starting from zero, in cm.
    *j*
      Array, shape (n, 4). Emission coefficients: ``j_{IQUV}``, in that order.
    *K*
      Array, shape (n, 7). Absorption coefficients and Faraday mixing coefficients:
      ``alpha_{IQUV}, rho_{QUV}``.
    Return value
      Array of shape (4, m): Stokes intensities ``IQUV`` along parts of the
      ray with non-zero total emissivities; m <= n.

    """
    # A correct version of the fully-specified absorption matrix is in Leung+
    # 2011 (10.1088/0004-637X/737/1/21). From looking at the function
    # `radtrans_jac_form` of grtrans' `radtrans_integrate.f90` file, one can
    # see that the K vector is indeed packed in the way described above.
    from radtrans_integrate import radtrans_integrate

    n = x.size

    # The formal method doesn't do the same kind of clipping as LSODA *inside*
    # grtrans, but here we do the same clipping for consistency, and since it
    # is genuinely true that the clipped samples are superfluous.

    if np.all(j[:,0] == 0.):
        return np.zeros((4, n))

    i0 = 0
    i1 = n - 1

    while j[i0,0] == 0.:
        i0 += 1
    while j[i1,0] == 0.:
        i1 -= 1

    n = i1 + 1 - i0
    x = x[i0:i1+1]
    j = j[i0:i1+1]
    K = K[i0:i1+1]

    # OK we can go.

    radtrans_integrate.init_radtrans_integrate_data(
        METHOD_FORMAL, # method selector
        4, # number of equations
        n, # number of input data points
        n, # number of output data points
        10., # maximum optical depth; not used by "formal"
        1., # maximum absolute step size; not used by "formal"
        0.1, # absolute tolerance; not used by "formal"
        0.1, # relative tolerance; not used by "formal"
        1e-2, # "thin" parameter for DELO method; not used by "formal"
        1, # maximum number of steps; not used by "formal"
    )

    try:
        tau = x # not used by "formal"
        radtrans_integrate.integrate(x[::-1], j[::-1], K[::-1], tau[::-1], 4)
        i = radtrans_integrate.intensity.copy()
    finally:
        # If we exit without calling this, the next init call causes an abort
        radtrans_integrate.del_radtrans_integrate_data()
    return i


def integrate_ray_lsoda(x, j, K, atol=1e-8, rtol=1e-6, max_step_size=None,
                        frac_max_step_size=1e-3, max_steps=100000):
    """Use `grtrans <https://github.com/jadexter/grtrans>`_ to integrate one ray
    using its LSODA method.

    **Call signature**

    *x*
      1D array, shape (n,). Path length along ray, starting from zero, in cm.
    *j*
      Array, shape (n, 4). Emission coefficients: ``j_{IQUV}``, in that order.
    *K*
      Array, shape (n, 7). Absorption coefficients and Faraday mixing coefficients:
      ``alpha_{IQUV}, rho_{QUV}``.
    *atol*
      Some kind of tolerance parameter.
    *rtol*
      Some kind of tolerance parameter.
    *max_step_size*
      The maximum absolute step size. Overrides *frac_max_step_size*.
    *frac_max_step_size*
      If *max_step_size* is not specified, the maximum step size passed to the
      integrator is ``x.max()`` multiplied by this parameter. Experience shows
      that (for LSODA at least) this parameter must be pretty small to get
      good convergence!
    *max_steps*
      The maximum number of steps to take.
    Return value
      Array of shape (4, m): Stokes intensities IQUV along parts of the ray with
      non-zero total emissivities; m <= n.

    """
    n = x.size

    if max_step_size is None:
        max_step_size = frac_max_step_size * x.max()

    # the LSODA method clips its input arrays based on "tau" and zero emission
    # coefficients. It's hard for us to find out how it clipped, though, so we
    # reproduce its logic. LSODA doesn't use "tau" for anything besides this
    # clipping, so we pass it all zeros.

    if np.all(j[:,0] == 0.):
        return np.zeros((4, n))

    i0 = 0
    i1 = n - 1

    while j[i0,0] == 0.:
        i0 += 1
    while j[i1,0] == 0.:
        i1 -= 1

    n = i1 + 1 - i0
    x = x[i0:i1+1]
    j = j[i0:i1+1]
    K = K[i0:i1+1]

    # OK we can go.

    radtrans_integrate.init_radtrans_integrate_data(
        METHOD_LSODA_YES_LINEAR_STOKES, # method selector
        4, # number of equations
        n, # number of input data points
        n, # number of output data points
        10., # maximum optical depth; defused here (see comment above)
        max_step_size, # maximum absolute step size
        atol, # absolute tolerance
        rtol, # relative tolerance
        1e-2, # "thin" parameter for DELO method ... to be researched
        max_steps, # maximum number of steps
    )

    try:
        tau = np.zeros(n)
        radtrans_integrate.integrate(x[::-1], j, K, tau, 4)
        i = radtrans_integrate.intensity.copy()
    finally:
        # If we exit without calling this, the next init call causes an abort
        radtrans_integrate.del_radtrans_integrate_data()
    return i


def integrate(d, coeffs, psi):
    """Integrate a ray with `grtrans <https://github.com/jadexter/grtrans>`_,
    using reasonable defaults.

    **Call signature**

    *d*
      An array giving the location of each sample along the ray, starting from zero, in cm.
    *coeffs*
      An array of shape (N, 8) of RT coefficients in the basis where the
      Stokes U coefficients are always zero. Such arrays are returned by
      :meth:`neurosynchro.impl.PhysicalApproximator.compute_all_nontrivial`.
    *psi*
      An array of angles between the local magnetic field and the observer’s Stokes U
      axis, in radians.
    Return value
      An array of shape (4,), giving the Stokes IQUV at the end of the ray.

    This function is mainly intended to test what happens if the passed-in
    coefficients are slightly different due to the neural network
    approximation. So we don't provide many knobs or diagnostics here.

    """
    from . import detrivialize_stokes_basis

    xformed = detrivialize_stokes_basis(coeffs, psi)
    j = np.empty(coeffs.shape[:-1] + (4,))
    j[...,0] = xformed[...,0]
    j[...,1] = xformed[...,2]
    j[...,2] = xformed[...,4]
    j[...,3] = xformed[...,6]
    K = np.empty(coeffs.shape[:-1] + (7,))
    K[...,0] = xformed[...,1]
    K[...,1] = xformed[...,3]
    K[...,2] = xformed[...,5]
    K[...,3:] = xformed[...,7:]

    iquv = integrate_ray_formal(d, j, K)
    return iquv[:,-1]


def make_parser(ap=None):
    if ap is None:
        ap = argparse.ArgumentParser()

    ap.add_argument('--frequency', '-f', type=float, metavar='<ghz>',
                    default=10.0, help='The observing frequency to simulate')
    ap.add_argument('nndir', type=str, metavar='<nndir>',
                    help='The path to the neural-net directory.')
    ap.add_argument('testdata', type=str, metavar='<testdata>',
                    help='A file containing test data')
    return ap


def grtrans_cli(settings):
    from pwkit import cgs
    from pwkit.cli import die
    from time import time
    from .impl import PhysicalApproximator, hardcoded_nu_ref, hardcoded_ne_ref

    # Read and validate the test dataset.

    testdata = pd.read_table(settings.testdata)

    psi = testdata.get('psi(meta)')
    if psi is None:
        die('the test dataset must contain a column of field-to-Stokes-U angles \"psi(meta)\"')

    d = testdata.get('d(meta)')
    if d is None:
        die('the test dataset must contain a column of integration path lengths \"d(meta)\"')

    n_e = testdata.get('n_e(meta)')
    if n_e is None:
        die('the test dataset must contain a column of particle densities \"n_e(meta)\"')

    time_ms = testdata.get('time_ms(meta)')
    if time_ms is None:
        die('the test dataset must contain a column of computation times \"time_ms(meta)\"')

    s = None
    theta = None
    others = {}

    for col in testdata.columns:
        if col.startswith('s('):
            s = testdata[col]
        elif col.startswith('theta('):
            theta = testdata[col]
        elif col.endswith('(lin)') or col.endswith('(log)'):
            others[col.split('(')[0]] = testdata[col]

    if s is None:
        die('the test dataset must have an input parameter of the harmonic number \"s\"')

    if theta is None:
        die('the test dataset must have an input parameter of the field-to-LOS angle \"theta\"')

    # Get the coefficients into physical units, packed in our standard format.

    nu_hz = settings.frequency * 1e9
    freq_scale = nu_hz / hardcoded_nu_ref
    n_e_scale = n_e / hardcoded_ne_ref

    coeffs = np.empty((psi.size, 8))
    coeffs[...,0] = testdata['j_I(res)'] * freq_scale
    coeffs[...,1] = testdata['alpha_I(res)'] / freq_scale
    coeffs[...,2] = testdata['j_Q(res)'] * freq_scale
    coeffs[...,3] = testdata['alpha_Q(res)'] / freq_scale
    coeffs[...,4] = testdata['j_V(res)'] * freq_scale
    coeffs[...,5] = testdata['alpha_V(res)'] / freq_scale
    coeffs[...,6] = testdata['rho_Q(res)'] / freq_scale
    coeffs[...,7] = testdata['rho_V(res)'] / freq_scale
    coeffs *= n_e_scale.values.reshape((-1, 1))

    # Ground truth:

    iquv_precise = integrate(d, coeffs, psi)
    ctime_precise = time_ms.sum()
    print('Precise computation: I={:.4e}  Q={:.4e}  U={:.4e}  V={:.4e}  calc_time={:.0f} ms'.format(
          iquv_precise[0], iquv_precise[1], iquv_precise[2], iquv_precise[3], ctime_precise
    ))

    # Now set up the approximator and do the same thing. (Note that often the
    # timing seems backwards, because the time spent doing the precise
    # calculation has already been spent, whereas we have a lot of overhead to
    # set up the neural networks.)

    B = 2 * np.pi * cgs.me * cgs.c * nu_hz / (s * cgs.e)
    approx = PhysicalApproximator(settings.nndir)
    t0 = time()
    coeffs, oos = approx.compute_all_nontrivial(nu_hz, B, n_e, theta, **others)
    ctime_approx = 1000 * (time() - t0)

    if np.any(oos != 0):
        print('WARNING: some of the approximations were out-of-sample')

    iquv_approx = integrate(d, coeffs, psi)
    print('Approx. computation: I={:.4e}  Q={:.4e}  U={:.4e}  V={:.4e}  calc_time={:.0f} ms'.format(
          iquv_approx[0], iquv_approx[1], iquv_approx[2], iquv_approx[3], ctime_approx
    ))

    acc = np.abs((iquv_approx - iquv_precise) / iquv_precise)
    print('Accuracy: I={:.3f}  Q={:.3f}  U={:.3f}  V={:.3f}'.format(acc[0], acc[1], acc[2], acc[3]))

    print('Speedup: {:.1f}'.format(ctime_precise / ctime_approx))



# -*- mode: python; coding: utf-8 -*-
# Copyright 2017-2018 Peter Williams and collaborators.
# Licensed under the MIT License.

"""Train one of the neural networks.

Meant to be run as a program in production, but you can import it to
experiment with training regimens.

"""
from __future__ import absolute_import, division, print_function

import argparse, sys, time
from pwkit.cli import die
from pwkit.io import Path

from . import DomainRange


def generic_trainer(m):
    from keras.layers import Dense

    m.add(Dense(
        units = 300,
        input_dim = m.domain_range.n_params,
        activation = 'relu',
        kernel_initializer = 'normal',
    ))
    m.add(Dense(
        units = 1,
        activation = 'linear',
        kernel_initializer = 'normal',
    ))
    m.compile('adam', 'mse')
    hist = m.ns_fit(
        epochs = 30,
        batch_size = 2000,
        verbose = 0,
    )
    print('Intermediate MSE:', hist.history['loss'][-1])
    m.ns_sigma_clip(7)
    hist = m.ns_fit(
        epochs = 30,
        batch_size = 2000,
        verbose = 0,
    )
    m.final_mse = hist.history['loss'][-1]
    return m


def twolayer_trainer(m):
    from keras.layers import Dense

    m.add(Dense(
        units = 120,
        input_dim = m.domain_range.n_params,
        activation = 'relu',
        kernel_initializer = 'normal',
    ))
    m.add(Dense(
        units = 60,
        activation = 'relu',
        kernel_initializer = 'normal',
    ))
    m.add(Dense(
        units = 1,
        activation = 'linear',
        kernel_initializer = 'normal',
    ))
    m.compile('adam', 'mse')
    hist = m.ns_fit(
        epochs = 30,
        batch_size = 2000,
        verbose = 0,
    )
    print('Intermediate MSE:', hist.history['loss'][-1])
    m.ns_sigma_clip(7)
    hist = m.ns_fit(
        epochs = 30,
        batch_size = 2000,
        verbose = 0,
    )
    m.final_mse = hist.history['loss'][-1]
    return m


def binary_trainer(m):
    from keras.layers import Dense

    m.add(Dense(
        units = 120,
        input_dim = m.domain_range.n_params,
        activation = 'relu',
        kernel_initializer = 'normal',
    ))
    m.add(Dense(
        units = 60,
        activation = 'relu',
        kernel_initializer = 'normal',
    ))
    m.add(Dense(
        units = 1,
        activation = 'sigmoid',
        kernel_initializer = 'normal',
    ))
    m.compile(optimizer='adam', loss='binary_crossentropy')
    hist = m.ns_fit(
        epochs = 30,
        batch_size = 2000,
        verbose = 0,
    )
    print('Intermediate MSE:', hist.history['loss'][-1])
    # Note: no sigma-clipping
    hist = m.ns_fit(
        epochs = 30,
        batch_size = 2000,
        verbose = 0,
    )
    m.final_mse = hist.history['loss'][-1]
    return m


def load_data_and_train(datadir, nndir, result_name):
    from .impl import NSModel

    cfg_path = Path(nndir) / 'nn_config.toml'
    dr, rinfo = DomainRange.from_serialized(cfg_path, result_to_extract=result_name)

    if rinfo is None:
        die('no known result named %r', result_name)

    sd = dr.load_and_normalize(datadir)

    trainer_name = rinfo['trainer']
    trainer_func = globals().get(trainer_name + '_trainer')
    if trainer_func is None:
        die('unknown trainer function %r', trainer_name)

    print('Training with scheme \"%s\"' % trainer_name)
    m = NSModel()
    m.ns_setup(rinfo['_index'], sd)
    t0 = time.time()
    trainer_func(m)
    m.training_wall_clock = time.time() - t0
    return m


def page_results(m, residuals=False, thin=500):
    import omega as om

    pg = om.makeDisplayPager()
    for i in range(m.domain_range.n_params):
        pg.send(m.ns_plot(i, plot_err=residuals, thin=thin))

    pg.done()


def make_parser(ap=None):
    if ap is None:
        ap = argparse.ArgumentParser()

    ap.add_argument('-p', '--plot', action='store_true',
                    help='Compare the NN and Symphony after training.')
    ap.add_argument('-r', '--residuals', action='store_true',
                    help='If plotting, plot residuals rather than absolute values (requires `omegaplot` package).')
    ap.add_argument('datadir', type=str, metavar='<datadir>',
                    help='The path to the sample data directory.')
    ap.add_argument('nndir', type=str, metavar='<nndir>',
                    help='The path to the neural-net directory.')
    ap.add_argument('result_name', type=str, metavar='<result-name>',
                    help='The name of the simulation result to train on.')
    return ap


def train_cli(settings):
    m = load_data_and_train(settings.datadir, settings.nndir, settings.result_name)
    print('Achieved MSE of %g in %.1f seconds for %s.' %
          (m.final_mse, m.training_wall_clock, settings.result_name))

    if settings.plot:
        page_results(m, residuals=settings.residuals)

    outpath = str(Path(settings.nndir) / ('%s.h5' % settings.result_name))
    m.save(outpath, overwrite=True)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2018 Peter Williams and collaborators
# Licensed under the MIT License

from __future__ import absolute_import, division, print_function

from setuptools import setup, Extension

# Keep synchronized with doc/conf.py:
version = '0.1.4'

setup(
    name = 'neurosynchro',
    version = version,
    author = 'Peter Williams',
    author_email = 'peter@newton.cx',
    url = 'https://github.com/pkgw/neurosynchro/',
    license = 'MIT',
    description = ('Use neural networks to approximate polarized synchrotron '
                   'radiative transfer coefficients'),

    long_description = '''\
*Neurosynchro* is a small Python package for creating and using neural
networks to quickly approximate the coefficients needed for fully-polarized
synchrotron radiative transfer. It builds on the `Keras <https://keras.io/>`_
deep learning library. Documentation may be found `on ReadTheDocs
<https://neurosynchro.readthedocs.io/en/stable/>`_.

Say that you have a code — such as `Rimphony
<https://github.com/pkgw/rimphony/>`_ or `Symphony
<https://github.com/AFD-Illinois/symphony>`_ — that calculates synchrotron
radiative transfer coefficients as a function of some input model parameters
(electron temperature, particle energy index, etc.). These calculations are
often accurate but slow. With *neurosynchro*, you can train a neural network
that will quickly approximate these calculations with good accuracy. The
achievable level of accuracy will depend on the particulars of your target
distribution function, range of input parameters, and so on.

This code is specific to synchrotron radiation because it makes certain
assumptions about how the coefficients scale with input parameters such as the
observing frequency.''',

    # Synchronize with README.md and doc/requirements.txt:
    install_requires = [
        'keras >=2.1',
        'numpy >=1.10',
        'pandas >=0.23.0',
        'pwkit >=0.8.19',
        'pytoml >=0.1.0',
        'six >=1.10',
    ],

    packages = [
        'neurosynchro',
    ],

    entry_points = {
        'console_scripts': [
            'neurosynchro = neurosynchro.cli:main',
        ],
    },

    include_package_data = True,

    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Topic :: Scientific/Engineering :: Astronomy',
    ],

    keywords = 'neural-networks radiative-transfer science',
)



#! /usr/bin/env python
# -*- mode: python; coding: utf-8 -*-
# Copyright 2017-2018 Peter Williams and collaborators.
# Licensed under the MIT License.

"""Command-line access to neurosynchro functionality.

"""
from __future__ import absolute_import, division, print_function

import argparse, sys
import numpy as np
from pwkit.cli import die
from pwkit.io import Path
import pytoml

from . import basic_load



def _hack_pytoml():
    """pytoml will stringify floats using repr, which is ugly and fails outright with
    very small values (i.e. 1e-30 becomes "0.000...."). Here we hack it to use
    exponential notation if needed.

    """
    from pytoml import writer
    orig_format_value = writer._format_value

    if not getattr(orig_format_value, '_neurosynchro_hack_applied', False):
        def better_format_value(v):
            if isinstance(v, float):
                if not np.isfinite(v):
                    raise ValueError("{0} is not a valid TOML value".format(v))
                return '%.16g' % v
            return orig_format_value(v)

        better_format_value._neurosynchro_hack_applied = True
        writer._format_value = better_format_value

_hack_pytoml()


# The "init-nndir" subcommand

def make_nninit_parser(ap=None):
    if ap is None:
        ap = argparse.ArgumentParser()

    ap.add_argument('nndir', type=str, metavar='<nndir>',
                    help='The name of the output neural-net directory to create')
    return ap


NNINIT_DEFAULT_CONFIG = {
    'params': [
        dict(
            name = 's',
            maptype = 'log',
        ),

        dict(
            name = 'theta',
            maptype = 'direct',
            phys_bounds_mode = 'theta',
            out_of_sample = 'clip',
        )
    ],

    'results': [
        dict(
            name = 'j_I',
            maptype = 'log',
            trainer = 'generic',
        ),

        dict(
            name = 'alpha_I',
            maptype = 'log',
            trainer = 'generic',
        ),

        dict(
            name = 'rho_Q',
            maptype = 'abs_log',
            trainer = 'generic',
        ),

        dict(
            name = 'rho_V',
            maptype = 'log',
            trainer = 'generic',
        ),

        dict(
            name = 'j_frac_pol',
            maptype = 'logit',
            trainer = 'generic',
        ),

        dict(
            name = 'alpha_frac_pol',
            maptype = 'logit',
            trainer = 'generic',
        ),

        dict(
            name = 'j_V_share',
            maptype = 'logit',
            trainer = 'generic',
        ),

        dict(
            name = 'alpha_V_share',
            maptype = 'logit',
            trainer = 'generic',
        ),

        dict(
            name = 'rho_Q_sign',
            maptype = 'direct',
            normalization_mode = 'unit_interval',
            x_mean = -1,
            x_std = 2,
            phys_min = -1,
            phys_max = 1,
            norm_min = 0,
            norm_max = 1,
        ),
    ]
}

def nninit_cli(settings):
    nndir = Path(settings.nndir)

    try:
        nndir.mkdir()
    except OSError as e:
        if e.errno == 17:
            die('directory \"%s\" already exists' % settings.nndir)
        raise

    cfg_path = nndir / 'nn_config.toml'
    with cfg_path.open('wt') as f:
        pytoml.dump(f, NNINIT_DEFAULT_CONFIG)


# The "lock-domain-range" subcommand

def make_ldr_parser(ap=None):
    if ap is None:
        ap = argparse.ArgumentParser()

    ap.add_argument('datadir', type=str, metavar='<datadir>',
                    help='The path to the input training data directory.')
    ap.add_argument('nndir', type=str, metavar='<nndir>',
                    help='The path to the output neural-net directory.')
    return ap


def lock_domain_range_cli(settings):
    from . import DomainRange

    # Load samples
    df = basic_load(settings.datadir)

    # Load skeleton config
    cfg_path = Path(settings.nndir) / 'nn_config.toml'
    with cfg_path.open('rt') as f:
        info = pytoml.load(f)

    # Turn into processed DomainRange object
    dr = DomainRange.from_info_and_samples(info, df)

    # Update config and rewrite
    dr.into_info(info)

    with cfg_path.open('wt') as f:
        pytoml.dump(f, info)


# The "summarize" subcommand

def summarize(datadir):
    df = basic_load(datadir)

    # Report stuff.

    print('Columns:', ' '.join(df.columns))
    print('Number of rows:', df.shape[0])
    print('Total number of NaNs:', np.isnan(df.values).sum())
    print('Number of rows with NaNs:', (np.isnan(df.values).sum(axis=1) > 0).sum())

    for c in df.columns:
        r = df[c]
        print()
        print('Column %s:' % c)
        print('  Number of NaNs:', np.isnan(r).sum())
        print('  Non-NaN max:', np.nanmax(r))
        print('  Non-NaN min:', np.nanmin(r))
        print('  Nonnegative:', (r >= 0).sum())
        print('  Nonpositive:', (r <= 0).sum())


def make_summarize_parser(ap=None):
    if ap is None:
        ap = argparse.ArgumentParser()

    ap.add_argument('datadir', type=str, metavar='<datadir>',
                    help='The path to the sample data directory.')
    return ap


def summarize_cli(settings):
    summarize(settings.datadir)


# The "transform" subcommand

def transform(datadir):
    """This task takes the raw synchrotron coefficients output by rimphony and
    transforms them into a format that better respects the physical
    constraints of the problem.

    """
    import pandas as pd

    df = basic_load(datadir)
    n = df.shape[0]

    df = df.dropna()
    print('Dropping due to NaNs:', n - df.shape[0], file=sys.stderr)

    bad = (df['j_I(res)'] <= 0)
    mask = bad
    print('Rows with bad J_I:', bad.sum(), file=sys.stderr)

    bad = (df['alpha_I(res)'] <= 0)
    mask |= bad
    print('Rows with bad a_I:', bad.sum(), file=sys.stderr)

    bad = (df['j_Q(res)'] >= 0)
    mask |= bad
    print('Rows with bad J_Q:', bad.sum(), file=sys.stderr)

    bad = (df['alpha_Q(res)'] >= 0)
    mask |= bad
    print('Rows with bad a_Q:', bad.sum(), file=sys.stderr)

    bad = (df['j_V(res)'] <= 0)
    mask |= bad
    print('Rows with bad J_V:', bad.sum(), file=sys.stderr)

    bad = (df['alpha_V(res)'] <= 0)
    mask |= bad
    print('Rows with bad a_V:', bad.sum(), file=sys.stderr)

    # This cut isn't physically motivated, but under the current rimphony
    # model, f_V is always positive.
    bad = (df['rho_V(res)'] <= 0)
    mask |= bad
    print('Rows with bad f_V:', bad.sum(), file=sys.stderr)

    n = df.shape[0]
    df = df[~mask]
    print('Dropped due to first-pass filters:', n - df.shape[0], file=sys.stderr)

    j_pol = np.sqrt(df['j_Q(res)']**2 + df['j_V(res)']**2)
    a_pol = np.sqrt(df['alpha_Q(res)']**2 + df['alpha_V(res)']**2)

    df['j_frac_pol(res)'] = j_pol / df['j_I(res)']
    bad = (df['j_frac_pol(res)'] < 0) | (df['j_frac_pol(res)'] > 1)
    mask = bad
    print('Rows with bad j_frac_pol:', bad.sum(), file=sys.stderr)

    df['alpha_frac_pol(res)'] = a_pol / df['alpha_I(res)']
    bad = (df['alpha_frac_pol(res)'] < 0) | (df['alpha_frac_pol(res)'] > 1)
    mask |= bad
    print('Rows with bad alpha_frac_pol:', bad.sum(), file=sys.stderr)

    n = df.shape[0]
    df = df[~mask]
    print('Dropped due to second-pass filters:', n - df.shape[0], file=sys.stderr)

    df['j_V_share(res)'] = df['j_V(res)'] / j_pol
    df['alpha_V_share(res)'] = df['alpha_V(res)'] / a_pol

    # I used to scale rho_{Q,V} by alpha_I, but these values are often
    # strongly different. (And, judging by the commentary in Heyvaerts, I
    # think this is probably OK and not a sign of a numerics problem.) So we
    # just pass those columns on through like {j,alpha}_I. However, in a bit
    # of a hack, we add a column giving the sign of the rho_Q column, since in
    # the "pitchy kappa" distribution we have a non-negligible number of
    # negative rho_Q values *plus* a large dynamic range on both sides of
    # zero. Adding this column lets us break the neural networking into two
    # pieces in a way that doesn't involve a bunch of complicated
    # rearchitecting of my parameter code.
    df['rho_Q_sign(res)'] = np.sign(df['rho_Q(res)'])

    print('Final row count:', df.shape[0], file=sys.stderr)

    for c in 'j_Q alpha_Q j_V alpha_V'.split():
        del df[c + '(res)']

    df.to_csv(
        sys.stdout,
        sep = '\t',
        index = False
    )


def make_transform_parser(ap=None):
    if ap is None:
        ap = argparse.ArgumentParser()

    ap.add_argument('datadir', type=str, metavar='<datadir>',
                    help='The path to the sample data directory.')
    return ap


def transform_cli(settings):
    transform(settings.datadir)


# The entrypoint

def entrypoint(argv):
    ap = argparse.ArgumentParser(prog='neurosynchro')
    subparsers = ap.add_subparsers(
        dest = 'subcommand',
        metavar = '<command>',
        help = 'The sub-command to invoke'
    )

    make_nninit_parser(subparsers.add_parser(
        'init-nndir',
        help = 'Initialize a directory to save the neural network training data'
    ))

    make_ldr_parser(subparsers.add_parser(
        'lock-domain-range',
        help = 'Find the domain and range of the training set'
    ))

    make_summarize_parser(subparsers.add_parser(
        'summarize',
        help = 'Print summary statistics about the training set'
    ))

    from .grtrans import make_parser as make_grtrans_parser
    make_grtrans_parser(subparsers.add_parser(
        'test-grtrans',
        help = 'Do a test integration with grtrans'
    ))

    make_transform_parser(subparsers.add_parser(
        'transform',
        help = 'Transform the training set into Neurosynchro\'s internal parametrization',
        description = 'Transform the training set into Neurosynchro\'s internal parametrization.',
        epilog = '''The training set can have arbitrary input parameters, but should have eight
output parameters named `j_I`, `j_Q`, `j_V`, `alpha_I`, `alpha_Q`, `alpha_V`,
`rho_Q`, `rho_V` -- these are the standard Stokes-basis radiative transfer
coefficients. The transformed training set will be printed to standard output,
so you almost surely want to redirect the output of this program to a file.'''
    ))

    from .train import make_parser as make_train_parser
    make_train_parser(subparsers.add_parser(
        'train',
        help = 'Train one of the neural networks'
    ))

    settings = ap.parse_args(argv[1:])

    if settings.subcommand is None:
        die('you must supply a subcommand; run with "--help" for help')

    if settings.subcommand == 'init-nndir':
        nninit_cli(settings)
    elif settings.subcommand == 'lock-domain-range':
        lock_domain_range_cli(settings)
    elif settings.subcommand == 'summarize':
        summarize_cli(settings)
    elif settings.subcommand == 'test-grtrans':
        from .grtrans import grtrans_cli
        grtrans_cli(settings)
    elif settings.subcommand == 'train':
        from .train import train_cli
        train_cli(settings)
    elif settings.subcommand == 'transform':
        transform_cli(settings)
    else:
        # argparse will error out if it the user gives an unrecognized
        # subcommand, so if we get here it's an internal bug
        assert False, 'internal bug: forgot to handle subcommand!'


def main():
    import sys
    from pwkit import cli

    cli.unicode_stdio()
    cli.propagate_sigint()
    cli.backtrace_on_usr1()
    entrypoint(sys.argv)
